# This file is part of dh-cmake, and is distributed under the OSI-approved
# BSD 3-Clause license. See top-level LICENSE file or
# https://gitlab.kitware.com/debian/dh-cmake/blob/master/LICENSE for details.

import json
import os.path
import re

from dhcmake import common


class DHCPack(common.DHCommon):
    def read_cpack_metadata(self):
        with open("debian/.cpack/cpack-metadata.json", "r") as f:
            self.cpack_metadata = json.load(f)

    def get_cpack_components(self, package):
        opened_file = self.read_package_file(package, "cpack-components")
        if opened_file:
            retval = []
            with opened_file as f:
                for l in f:
                    if not re.search("^($|#)", l):
                        group = l.rstrip()
                        if group in self.cpack_metadata["components"]:
                            retval.append(group)
                        else:
                            raise ValueError(
                                "Invalid CPack components in %s" % package)
            return retval
        else:
            return []

    def get_cpack_component_groups(self, package):
        opened_file = self.read_package_file(package, "cpack-component-groups")
        if opened_file:
            retval = []
            with opened_file as f:
                for l in f:
                    if not re.search("^($|#)", l):
                        group = l.rstrip()
                        if group in self.cpack_metadata["componentGroups"]:
                            retval.append(group)
                        else:
                            raise ValueError(
                                "Invalid CPack component groups in %s" %
                                package)
            return retval
        else:
            return []

    def get_all_cpack_components_for_group(self, group, visited=None):
        if visited is None:
            visited = set()

        if group in visited:
            return set()
        visited.add(group)

        all_components = set(self.cpack_metadata["componentGroups"][group]
                             ["components"])

        for sub_group in self.cpack_metadata["componentGroups"][group]["subgroups"]:
            all_components.update(
                self.get_all_cpack_components_for_group(
                    sub_group, visited))

        return all_components

    def get_all_cpack_components(self, package):
        all_components = set(self.get_cpack_components(package))

        for group in self.get_cpack_component_groups(package):
            all_components.update(
                self.get_all_cpack_components_for_group(group))

        return all_components

    def get_package_dependencies(self, package):
        deps = set()

        for component in self.get_all_cpack_components(package):
            for component_dep in self.cpack_metadata["components"][component]["dependencies"]:
                for other_package in self.get_packages():
                    if component_dep in \
                            self.get_all_cpack_components(other_package):
                        deps.add(other_package)

        return deps

    @common.DHEntryPoint("dh_cpack_generate")
    def generate(self, args=None):
        self.parse_args(args)

        cmd_args = [
            "cpack",
            "--config",
            os.path.join(self.get_build_directory(), "CPackConfig.cmake"),
            "-G", "External",
            "-D", "CPACK_PACKAGE_FILE_NAME=cpack-metadata",
            "-D", "CPACK_EXT_REQUESTED_VERSIONS=1.0",
            "-B", "debian/.cpack",
        ]
        self.do_cmd(cmd_args)

    @common.DHEntryPoint("dh_cpack_substvars")
    def substvars(self, args=None):
        self.parse_args(args)
        self.read_cpack_metadata()

        for package in self.get_packages():
            depends = ", ".join(dep + " (= ${binary:Version})" for dep in
                                sorted(self.get_package_dependencies(package)))
            if depends:
                self.write_substvar("cpack:Depends", depends, package)

    def install_make_arg_parser(self, parser):
        self.make_arg_parser(parser)
        parser.add_argument(
            "--sourcedir", action="store",
            help="Source directory for installation (not used except to notify"
                 " dh_missing)",
            default="debian/tmp")

    @common.DHEntryPoint("dh_cpack_install")
    def install(self, args=None):
        self.parse_args(args, make_arg_parser=self.install_make_arg_parser)
        self.read_cpack_metadata()

        for package in self.get_packages():
            for component in self.get_all_cpack_components(package):
                for project in self.cpack_metadata["projects"]:
                    if component in project["components"]:
                        extra_args = []

                        try:
                            extra_args.extend([
                                "--config",
                                self.cpack_metadata["buildType"]
                            ])
                        except KeyError:
                            pass

                        # TODO Fix this in CMake (https://gitlab.kitware.com/cmake/cmake/-/issues/20700)
                        # try:
                        #    extra_args.append(
                        #            "-DCMAKE_INSTALL_DEFAULT_"
                        #            "DIRECTORY_PERMISSIONS:STRING=" +
                        #            self.cpack_metadata[
                        #                "defaultDirectoryPermissions"])
                        # except KeyError:
                        #    pass

                        if self.cpack_metadata["stripFiles"]:
                            extra_args.append("--strip")

                        self.do_cmake_install(
                            project["directory"], package,
                            component=component,
                            extra_args=extra_args)


def generate():
    dhcpack = DHCPack()
    dhcpack.generate()


def substvars():
    dhcpack = DHCPack()
    dhcpack.substvars()


def install():
    dhcpack = DHCPack()
    dhcpack.install()



# This file is part of dh-cmake, and is distributed under the OSI-approved
# BSD 3-Clause license. See top-level LICENSE file or
# https://gitlab.kitware.com/debian/dh-cmake/blob/master/LICENSE for details.

import contextlib
import os
from dhcmake import cpack, arch
from . import DebianSourcePackageTestCaseBase, KWTestCaseBase

from debian import debfile, deb822


class DHCPackTestCase(DebianSourcePackageTestCaseBase):
    DHClass = cpack.DHCPack

    libraries_files = set(KWTestCaseBase.replace_arch_in_paths({
        "usr",
        "usr/lib",
        "usr/lib/{arch}",
        "usr/lib/{arch}/libdh-cmake-test.so.1",
        "usr/lib/{arch}/libdh-cmake-test.so.1.0",
        "usr/lib/{arch}/libdh-cmake-test-lib1.so.1",
        "usr/lib/{arch}/libdh-cmake-test-lib1.so.1.0",
        "usr/lib/{arch}/libdh-cmake-test-lib2.so.1",
        "usr/lib/{arch}/libdh-cmake-test-lib2.so.1.0",
    }))

    shlibs_files = {
        "DEBIAN",
        "DEBIAN/shlibs",
    }

    headers_files = set(KWTestCaseBase.replace_arch_in_paths({
        "usr",
        "usr/include",
        "usr/include/dh-cmake-test.h",
        "usr/include/dh-cmake-test-lib1.h",
        "usr/include/dh-cmake-test-lib2.h",
    }))

    namelinks_files = set(KWTestCaseBase.replace_arch_in_paths({
        "usr",
        "usr/lib",
        "usr/lib/{arch}",
        "usr/lib/{arch}/libdh-cmake-test.so",
        "usr/lib/{arch}/libdh-cmake-test-lib1.so",
        "usr/lib/{arch}/libdh-cmake-test-lib2.so",
    }))

    libdh_cmake_test_files = {
        "usr",
        "usr/share",
        "usr/share/doc",
        "usr/share/doc/libdh-cmake-test",
        "usr/share/doc/libdh-cmake-test/changelog.Debian.gz",
    }

    libdh_cmake_test_dev_files = {
        "usr",
        "usr/share",
        "usr/share/doc",
        "usr/share/doc/libdh-cmake-test-dev",
        "usr/share/doc/libdh-cmake-test-dev/changelog.Debian.gz",
    }

    def setUp(self):
        super().setUp()

        self.dh.parse_args([])

        os.mkdir(self.dh.get_build_directory())

        self.run_cmd(
            [
                "cmake", "-G", "Unix Makefiles", "-DCMAKE_INSTALL_PREFIX=/usr",
                self.src_dir,
            ], cwd=self.dh.get_build_directory())

        self.run_cmd(["make"], cwd=self.dh.get_build_directory())

    def test_generate(self):
        self.assertFileNotExists("debian/.cpack/cpack-metadata.json")
        self.dh.generate([])
        self.assertFileExists("debian/.cpack/cpack-metadata.json")

    def test_get_cpack_components(self):
        with open("debian/libdh-cmake-test-extra-32.cpack-components", "w") \
                as f:
            f.write("InvalidComponent\n")

        self.dh.generate([])
        self.dh.read_cpack_metadata()

        self.assertEqual(
            ["Libraries"],
            self.dh.get_cpack_components("libdh-cmake-test")
        )

        with self.assertRaises(
                ValueError,
                msg="Invalid CPack components in libdh-cmake-test-extra-32"
        ):
            self.dh.get_cpack_components("libdh-cmake-test-extra-32")

    def test_get_cpack_component_groups(self):
        with open("debian/libdh-cmake-test-extra-32.cpack-component-groups",
                  "w") as f:
            f.write("InvalidGroup\n")

        self.dh.generate([])
        self.dh.read_cpack_metadata()

        self.assertEqual(
            ["Development"],
            self.dh.get_cpack_component_groups("libdh-cmake-test-dev")
        )

        with self.assertRaises(
                ValueError,
                msg="Invalid CPack component groups "
                    "in libdh-cmake-test-extra-32"
        ):
            self.dh.get_cpack_component_groups("libdh-cmake-test-extra-32")

    def test_get_all_cpack_components_for_group(self):
        self.dh.generate([])
        self.dh.read_cpack_metadata()

        self.assertEqual({"Headers", "Namelinks"},
                         self.dh.get_all_cpack_components_for_group("Development"))

        self.assertEqual({"Libraries", "Headers", "Namelinks"},
                         self.dh.get_all_cpack_components_for_group("All"))

    def test_get_all_cpack_components(self):
        self.dh.generate([])
        self.dh.read_cpack_metadata()

        self.assertEqual({"Headers", "Namelinks"},
                         self.dh.get_all_cpack_components("libdh-cmake-test-dev"))

        self.assertEqual({"Libraries"},
                         self.dh.get_all_cpack_components("libdh-cmake-test"))

    def test_get_package_dependencies(self):
        self.dh.generate([])
        self.dh.read_cpack_metadata()

        self.assertEqual({"libdh-cmake-test"},
                         self.dh.get_package_dependencies("libdh-cmake-test-dev"))

    def test_get_package_dependencies_packages(self):
        self.dh.generate(["--package", "libdh-cmake-test-dev"])
        self.dh.read_cpack_metadata()

        self.assertEqual(set(),
                         self.dh.get_package_dependencies("libdh-cmake-test-dev"))

    def test_substvars(self):
        self.dh.generate([])
        self.dh.substvars([])

        with open("debian/libdh-cmake-test-dev.substvars", "r") as f:
            self.assertEqual("cpack:Depends=libdh-cmake-test "
                             "(= ${binary:Version})\n", f.read())

    def test_substvars_packages(self):
        self.dh.generate([])
        self.dh.substvars(["--package", "libdh-cmake-test-dev"])

        self.assertFileNotExists("debian/libdh-cmake-test-dev.substvars")

    def test_install(self):
        self.dh.generate([])
        self.dh.install([])

        self.assertFileTreeEqual(self.libraries_files,
                                 "debian/libdh-cmake-test")

        self.assertFileTreeEqual(self.headers_files | self.namelinks_files,
                                 "debian/libdh-cmake-test-dev")

    def test_run_debian_rules(self):
        self.run_debian_rules("build", "cpack")
        self.run_debian_rules("install", "cpack")

        self.assertFileTreeEqual(self.libraries_files | self.shlibs_files
                                 | self.libdh_cmake_test_files,
                                 "debian/libdh-cmake-test")

        self.assertFileTreeEqual(self.headers_files | self.namelinks_files
                                 | self.libdh_cmake_test_dev_files,
                                 "debian/libdh-cmake-test-dev")

        self.run_debian_rules("binary", "cpack")

        with contextlib.closing(debfile.DebFile(
            "../libdh-cmake-test-dev_0.1-1_%s.deb" %
                arch.dpkg_architecture()["DEB_HOST_ARCH"])) as f:
            packages = deb822.Packages(f.debcontrol())

            self.assertEqual([
                [
                    {
                        "name": "libdh-cmake-test",
                        "archqual": None,
                        "version": ("=", "0.1-1"),
                        "arch": None,
                        "restrictions": None,
                    },
                ],
            ], packages.relations["depends"])



# This file is part of dh-cmake, and is distributed under the OSI-approved
# BSD 3-Clause license. See top-level LICENSE file or
# https://gitlab.kitware.com/debian/dh-cmake/blob/master/LICENSE for details.

import errno
import os.path
import shutil
import subprocess
import sys
import tempfile
from unittest import TestCase

import dhcmake.common
import dhcmake.arch


class VolatileNamedTemporaryFile:
    def __init__(self, *args, **kwargs):
        self.ntf = tempfile.NamedTemporaryFile(*args, **kwargs)

    def close(self):
        try:
            self.ntf.close()
        except OSError as e:
            if e.errno != errno.ENOENT:
                raise

    @property
    def name(self):
        return self.ntf.name

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()


class KWTestCaseBase(TestCase):
    @classmethod
    def setUpClass(cls):
        cls.dh_verbose = os.environ.get("DH_VERBOSE", "") == "1"
        cls.stdout = sys.stdout
        cls.stderr = sys.stderr

        if not cls.dh_verbose:
            cls.stdout = open(os.devnull, "w")
            cls.stderr = open(os.devnull, "w")

    @classmethod
    def tearDownClass(cls):
        if not cls.dh_verbose:
            cls.stdout.close()
            cls.stderr.close()

    def assertFileExists(self, path):
        self.assertTrue(os.path.exists(
            path), "File '{0}' does not exist".format(path))

    def assertFileNotExists(self, path):
        self.assertFalse(os.path.exists(
            path), "File '{0}' exists".format(path))

    def assertFileContentsEqual(self, expected_contents, path):
        with open(path) as f:
            self.assertEqual(expected_contents, f.read())

    def assertFileTreeEqual(self, expected_files, path):
        actual_files = set()
        for dirpath, dirnames, filenames in os.walk(path):
            rel = os.path.relpath(dirpath, path)
            if rel == ".":
                actual_files.update(dirnames)
                actual_files.update(filenames)
            else:
                actual_files.update(os.path.join(rel, p) for p in dirnames)
                actual_files.update(os.path.join(rel, p) for p in filenames)

        self.assertEqual(expected_files, actual_files)

    def assertVolatileFileNotExists(self, name):
        try:
            self.assertFileNotExists(name)
        except AssertionError:
            os.unlink(name)
            raise

    @classmethod
    def replace_arch_in_paths(cls, paths):
        return (p.format(arch=dhcmake.arch
                         .dpkg_architecture()["DEB_HOST_MULTIARCH"])
                for p in paths)

    def get_single_element(self, l):
        self.assertEqual(1, len(l))
        return l[0]

    @classmethod
    def run_cmd(cls, args, cwd=None, env=None):
        subprocess.run(args, cwd=cwd, env=env, check=True,
                       stdout=cls.stdout, stderr=cls.stderr)


class DebianSourcePackageTestCaseBase(KWTestCaseBase):
    @classmethod
    def push_path(cls, name, value):
        try:
            old = os.environ[name]
        except KeyError:
            old = None

        if old is None:
            os.environ[name] = value
        else:
            os.environ[name] = value + ":" + old

        return old

    @classmethod
    def pop_path(cls, name, value):
        if value is None:
            del os.environ[name]
        else:
            os.environ[name] = value

    @classmethod
    def setUpClass(cls):
        super().setUpClass()

        test_dir = os.path.dirname(os.path.abspath(__file__))
        root_dir = os.path.dirname(os.path.dirname(test_dir))

        cls.install_dir = tempfile.TemporaryDirectory()

        cls.run_cmd([os.path.join(root_dir, "setup.py"), "install_scripts",
                     "--install-dir=" + os.path.join(cls.install_dir.name, "bin")])
        cls.run_cmd([os.path.join(root_dir, "setup.py"), "install_data",
                     "--install-dir=" + cls.install_dir.name])

        cls.old_path = cls.push_path(
            "PATH", os.path.join(cls.install_dir.name, "bin"))
        cls.old_perl5lib = cls.push_path(
            "PERL5LIB", os.path.join(cls.install_dir.name, "share/perl5"))
        cls.old_pythonpath = cls.push_path("PYTHONPATH", root_dir)

    @classmethod
    def tearDownClass(cls):
        cls.pop_path("PYTHONPATH", cls.old_pythonpath)
        cls.pop_path("PERL5LIB", cls.old_perl5lib)
        cls.pop_path("PATH", cls.old_path)

        cls.install_dir.cleanup()

        super().tearDownClass()

    def setUp(self):
        self.dh = self.DHClass()
        self.dh.stdout = self.stdout
        self.dh.stderr = self.stderr

        test_dir = os.path.dirname(os.path.abspath(__file__))
        test_data_dir = os.path.join(test_dir, "data")
        debian_pkg_dir = os.path.join(test_data_dir, "debian_pkg")

        self.tmp_dir = tempfile.TemporaryDirectory()
        self.src_dir = os.path.join(self.tmp_dir.name, "src")

        shutil.copytree(debian_pkg_dir, self.src_dir)

        self.old_cwd = os.getcwd()
        os.chdir(self.src_dir)

    def tearDown(self):
        os.chdir(self.old_cwd)

        self.tmp_dir.cleanup()

    def make_directory_in_tmp(self, name):
        path = os.path.join(self.tmp_dir.name, name)
        os.makedirs(path)
        return path

    def run_debian_rules(self, rule, case=None):
        env = os.environ.copy()
        if case:
            env["DH_CMAKE_CASE"] = case

        self.run_cmd(["debian/rules", rule], env=env)


class VolatileNamedTemporaryFileTestCase(KWTestCaseBase):
    def test_normal_delete(self):
        with VolatileNamedTemporaryFile() as f:
            pass
        self.assertVolatileFileNotExists(f.name)

    def test_already_deleted(self):
        with VolatileNamedTemporaryFile() as f:
            os.unlink(f.name)
        self.assertVolatileFileNotExists(f.name)



# This file is part of dh-cmake, and is distributed under the OSI-approved
# BSD 3-Clause license. See top-level LICENSE file or
# https://gitlab.kitware.com/debian/dh-cmake/blob/master/LICENSE for details.



# This file is part of dh-cmake, and is distributed under the OSI-approved
# BSD 3-Clause license. See top-level LICENSE file or
# https://gitlab.kitware.com/debian/dh-cmake/blob/master/LICENSE for details.


import subprocess


_known_archs = dict()


def debarch_is(real, alias):
    global _known_archs
    try:
        result = _known_archs[(real, alias)]
    except KeyError:
        result = subprocess.run(
            ["dpkg-architecture", "-i", alias, "-a", real, "-f"],
            stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL
        ).returncode == 0
        _known_archs[(real, alias)] = result
    return result


def debarch_contains(real, aliases):
    for alias in aliases:
        if debarch_is(real, alias):
            return True
    return False


_dpkg_architecture_values = None


def dpkg_architecture():
    global _dpkg_architecture_values
    if _dpkg_architecture_values is None:
        _dpkg_architecture_values = dict()
        proc = subprocess.run(["dpkg-architecture"], stdout=subprocess.PIPE)
        output = proc.stdout.decode()
        for line in output.split("\n"):
            if line:
                key, value = line.split("=", maxsplit=1)
                _dpkg_architecture_values[key] = value

    return _dpkg_architecture_values



# This file is part of dh-cmake, and is distributed under the OSI-approved
# BSD 3-Clause license. See top-level LICENSE file or
# https://gitlab.kitware.com/debian/dh-cmake/blob/master/LICENSE for details.

import os.path

from dhcmake import deb822
from . import KWTestCaseBase


class Deb822TestCase(KWTestCaseBase):
    def test_control(self):
        test_dir = os.path.dirname(os.path.abspath(__file__))
        test_data_dir = os.path.join(test_dir, "data")

        with open(os.path.join(test_data_dir, "debian_pkg/debian/control"),
                  "r") as f:
            source, packages = deb822.read_control(f)

        self.assertEqual("dh-cmake-test", source["source"])

        self.assertEqual(6, len(packages))

        package = packages[0]
        self.assertEqual("libdh-cmake-test", package["package"])
        self.assertEqual(["any"], package.architecture)

        package = packages[1]
        self.assertEqual("libdh-cmake-test-dev", package["package"])
        self.assertEqual(["any"], package.architecture)

        package = packages[2]
        self.assertEqual("libdh-cmake-test-doc", package["package"])
        self.assertEqual(["all"], package.architecture)

        package = packages[3]
        self.assertEqual("libdh-cmake-test-extra-32", package["package"])
        self.assertEqual(["armhf"], package.architecture)

        package = packages[4]
        self.assertEqual("libdh-cmake-test-extra-64", package["package"])
        self.assertEqual(["arm64"], package.architecture)

        package = packages[5]
        self.assertEqual("libdh-cmake-test-extra-both", package["package"])
        self.assertEqual(["armhf", "arm64"], package.architecture)



# This file is part of dh-cmake, and is distributed under the OSI-approved
# BSD 3-Clause license. See top-level LICENSE file or
# https://gitlab.kitware.com/debian/dh-cmake/blob/master/LICENSE for details.

import re

import debian.deb822


def read_control(sequence, *args, **kwargs):
    iterator = debian.deb822.Deb822.iter_paragraphs(sequence, *args, **kwargs)
    source = ControlSource(next(iterator).dump())
    packages = [ControlPackage(p.dump()) for p in iterator]

    return source, packages


class ControlSource(debian.deb822.Deb822):
    pass


class ControlPackage(debian.deb822.Deb822):
    @property
    def architecture(self):
        return re.split("\\s+", self["architecture"])



#!/usr/bin/env python3

# This file is part of dh-cmake, and is distributed under the OSI-approved
# BSD 3-Clause license. See top-level LICENSE file or
# https://gitlab.kitware.com/debian/dh-cmake/blob/master/LICENSE for details.

from setuptools import setup
import debian.changelog
import os.path


def get_current_version():
    root_dir = os.path.dirname(os.path.abspath(__file__))
    changelog_file = os.path.join(root_dir, "debian/changelog")
    with open(changelog_file) as f:
        changelog = debian.changelog.Changelog(f)
    return str(changelog.version)


def debian_version_to_python_version(version_str):
    if version_str[-1] == "~":
        return version_str[:-1] + ".dev1"
    return version_str


setup(
    name="dh-cmake",
    version=debian_version_to_python_version(get_current_version()),
    description="Debhelper program for CMake projects",
    url="https://gitlab.kitware.com/debian/dh-cmake",
    author="Kyle Edwards",
    author_email="kyle.edwards@kitware.com",
    maintainer="Kitware Debian Maintainers",
    maintainer_email="debian@kitware.com",
    classifiers=[
        "License :: OSI Approved :: BSD License",
    ],
    packages=["dhcmake"],
    test_suite="dhcmake.tests",
    install_requires=["python-debian"],
    entry_points={
        "console_scripts": [
            "dh_cmake_install=dhcmake.cmake:install",
            "dh_ctest_clean=dhcmake.ctest:clean",
            "dh_ctest_start=dhcmake.ctest:start",
            "dh_ctest_update=dhcmake.ctest:update",
            "dh_ctest_configure=dhcmake.ctest:configure",
            "dh_ctest_build=dhcmake.ctest:build",
            "dh_ctest_test=dhcmake.ctest:test",
            "dh_ctest_submit=dhcmake.ctest:submit",
            "dh_cpack_generate=dhcmake.cpack:generate",
            "dh_cpack_substvars=dhcmake.cpack:substvars",
            "dh_cpack_install=dhcmake.cpack:install",
        ],
    },
    package_data={
        "dhcmake": ["dh_ctest_driver.cmake"],
    },
    data_files=[
        ("share/perl5/Debian/Debhelper/Sequence", [
            "sequence/cmake.pm",
            "sequence/ctest.pm",
            "sequence/cpack.pm",
        ]),
        ("share/doc/dh-cmake", [
            "README.md",
        ]),
    ],
)
